#!/usr/bin/env python
#
#    imap2mbox.py v1.4.1
#
#    Copyright (C) 2008/2009 Michele <o-zone@zerozone.it> Pinassi
#
#    This software is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This software is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    1/9/2008 - First public release
#    29/10/2008 (v1.1.0) Added patches from Gregory Boyce <gboyce@badbelly.com> (OptionParser, SSL, read-only, skip not-matched) and some small fixes by me
#    18/11/2008 (v1.2.0) New patches from Gregory Boyce <gboyce@badbelly.com> that adds support for Exchange. Added compatibility with Python 2.3
#    13/03/2009 (v1.2.1) Added fixes on password dialog from Peter Huewe <PeterHuewe@gmx.de>
#    26/03/2009 (v1.3.0) Added new options FromDate, ToDate and MarkDelete
#    21/04/2009 (v1.3.1) Some minor fixes
#    24/08/2009 (v1.4.0) Added (beta) HTML export of mailbox
#    01/09/2009 (v1.4.1) Fixes GMail issue and add -v option

import getpass, imaplib, os, email, sys, getopt, re
from optparse import OptionParser
from datetime import tzinfo, timedelta, datetime
from email.utils import getaddresses
import mimetypes

if sys.version_info[0] >= 2 and sys.version_info[1] >= 5:
    # do stuff for Python 2.5
    from email.utils import unquote
else:
    from rfc822 import unquote

from fnmatch import *

# Parse commandline options
parser = OptionParser()
parser.add_option("-m", "--mailsrv", default=None,help="Mailserver to connect to")
parser.add_option("-p", "--port", type="int",help="Port to connect to")
parser.add_option("-u", "--username", default=getpass.getuser(), help="Username (defaults to current user)")
parser.add_option("-w", "--password", default=None, help="Specify Password (less secure)")
parser.add_option("-f", "--folder", default="*", help="Select folders to download (wildcards supported)")
parser.add_option("-s", "--ssl", action="store_true", help="Enable SSL")		  
parser.add_option("-v", "--verbose", action="store_true", help="Be verbose (debug)")		  
parser.add_option("-r", "--readonly", action="store_true", help="Do not mark messages as read")		  
parser.add_option("-t", "--test", action="store_true", help="Test Mode")
parser.add_option("-o", "--output", default="mbox", help="How to save mails ? mbox (default) or html (save text and mime attachments)")
parser.add_option("", "--fromdate", default=None, help="Start date (format: YYYY-MM-DD)")
parser.add_option("", "--todate", default=None, help="To date (format: YYYY-MM-DD)")
parser.add_option("", "--markdelete", action="store_true", help="Mark as deleted all saved messages")		  

(options, args) = parser.parse_args()

if options.mailsrv == None:
    parser.error("Argument 'mailsrv' missing.")

if options.password == None:
    getpass.getpass()

# Parse list
def parseList(listEntry):
    parser = re.compile('(\(.*\)) (".*") (.*)')
    return parser.match(mbox).groups() 

IMAPHOST = options.mailsrv
IMAPUSER = options.username

# Vai !

class progressBar:
	def __init__(self, minValue = 0, maxValue = 10, totalWidth=12):
		self.progBar = "[]"   # This holds the progress bar string
		self.min = minValue
		self.max = maxValue
		self.span = maxValue - minValue
		self.width = totalWidth
		self.amount = 0       # When amount == max, we are 100% done 
		self.updateAmount(0)  # Build progress bar string

	def updateAmount(self, newAmount = 0):
		if newAmount < self.min: newAmount = self.min
		if newAmount > self.max: newAmount = self.max
		self.amount = newAmount

		# Figure out the new percent done, round to an integer
		diffFromMin = float(self.amount - self.min)
		percentDone = (diffFromMin / float(self.span)) * 100.0
		percentDone = round(percentDone)
		percentDone = int(percentDone)

		# Figure out how many hash bars the percentage should be
		allFull = self.width - 2
		numHashes = (percentDone / 100.0) * allFull
		numHashes = int(round(numHashes))

		# build a progress bar with hashes and spaces
		self.progBar = "[" + '#'*numHashes + ' '*(allFull-numHashes) + "]"

		# figure out where to put the percentage, roughly centered
		percentPlace = (len(self.progBar) / 2) - len(str(percentDone)) 
		percentString = str(percentDone) + "%"

		# slice the percentage into the bar
		self.progBar = self.progBar[0:percentPlace] + percentString + self.progBar[percentPlace+len(percentString):]

	def __str__(self):
		return str(self.progBar)

try:
    # If SSL...
    if options.ssl:
	if options.verbose:
	    print "VERBOSE: Connecting to %s:%s using SSL..."%(IMAPHOST,options.port)
	if options.port:
	    imap4 = imaplib.IMAP4_SSL(IMAPHOST,options.port)
	else:
	    imap4 = imaplib.IMAP4_SSL(IMAPHOST)
    # else normal...
    else:
	if options.verbose:
	    print "VERBOSE: Connecting to %s:%s..."%(IMAPHOST,options.port)
	if options.port:
	    imap4 = imaplib.IMAP4(IMAPHOST,options.port)
	else:
	    imap4 = imaplib.IMAP4(IMAPHOST)
    # Let's go !
    if imap4:
	if options.verbose:
	    print "VERBOSE: Logging to %s:%s..."%(IMAPUSER,options.password)

	imap4.login(IMAPUSER, options.password)
	if options.readonly:
	    imap4.select(readonly=True)
	else:
	    imap4.select()
    else:
	print "IMAP4 Error"
	sys.exit()
except Exception,e:
    print "IMAP4 Error: %r\n"%(e)
    sys.exit()

if options.test:
    ret, mboxes = imap4.list()
    for mbox in mboxes:
        hasChild, dot, mbName = parseList(mbox)
	mbName = unquote(mbName)
	try:
	    if options.readonly:
		ret, numMex = imap4.select(mbName,readonly=True)
	    else:
		ret, numMex = imap4.select(mbName)
	    print "%s has %d messages"%(mbName,int(numMex[0]))
	except Exception,e:
	    print "Error: %r\n"%e
	    
    sys.exit()

userFromDate = datetime(1980,1,1)
userToDate = datetime.now()

if options.fromdate:
    userDate = options.fromdate.split('-')
    userFromDate = datetime(int(userDate[0]),int(userDate[1]),int(userDate[2]))

if options.todate:
    userDate = options.todate.split('-')
    userToDate = datetime(int(userDate[0]),int(userDate[1]),int(userDate[2]))

print "Saving all messages from %s to %s"%(userFromDate,userToDate)    

messageSaved = 0
messageSkipped = 0

try:
    os.mkdir(IMAPHOST)
except:
    pass
os.chdir(IMAPHOST)
try:
    os.mkdir(IMAPUSER)
except:
    pass
os.chdir(IMAPUSER)
# Inizia ad importare la casella postale
ret, mboxes = imap4.list()
for mbox in mboxes:
    hasChild, dot, mbName = parseList(mbox)
    mbName = unquote(mbName)
    if options.readonly:
	ret, numMex = imap4.select(mbName,readonly=True)
    else:
	ret, numMex = imap4.select(mbName)
    if options.verbose:
	print "VERBOSE: %s (%r) has %r messages"%(mbName,ret,numMex[0])
    if ret == 'OK' and fnmatch(mbName, options.folder):	    
	if int(numMex[0]) > 0:
            mbName = re.sub("/",".",mbName)
	    if os.path.isfile("%s.mbox"%(mbName)):
		print "%s.mbox exists. Skipping !"%(mbName)
	    else:
		if options.output == "html":
		    # Se salva in HTML crea directory con nome mbox...
		    try:
			os.mkdir(mbName)
		    except:
			print "Warning: directory %s already exists !"%(mbName)
		    htmlMexList = "<html><head><title>Messages into '%s'</title></head><body><ul>"%(mbName);
		else: 
		    # ...altrimenti crea file MBOX !
		    mboxFd = os.open("%s.mbox"%(mbName),os.O_WRONLY | os.O_CREAT)
		numMex = int(numMex[0])
		print "Running on %s (%d messages):"%(mbName,numMex)
		ret, data = imap4.search(None, 'ALL')
		pBar = progressBar(0,numMex,20)
		for num in data[0].split():
		    # Preleva messaggio
		    ret, data = imap4.fetch(num, '(RFC822)')
    		    msgData = email.message_from_string(data[0][1])    	    
    		    msgDate = email.Utils.parsedate(msgData.get('Date'))
    		    try:
    			tempFromDate = datetime(msgDate[0],msgDate[1],msgDate[2])
    		    except Exception,e:
    			print "Date error: %s"%e
    			messageSkipped = messageSkipped + 1
    			
    		    if(tempFromDate == None):
    			print "Date error: %s"%msgDate
    			messageSkipped = messageSkipped + 1
    		    elif(userFromDate < tempFromDate < userToDate):
		        # Scrivi MESSAGGIO
		        if options.output == "html":
		    	    # Azzera variabili
			    htmlAttachs = ""
			    htmlBody = ""			    
			    ext = ""
			    # Decodifica mail in HTML e salva eventuali allegati
			    mboxFd = os.open(os.path.join(mbName,"message%04d.html"%(int(num))),os.O_WRONLY | os.O_CREAT)
			    counter = 1
			    # Salva allegati
			    for part in msgData.walk():
    				# multipart/* are just containers
    				if part.get_content_maintype() == 'multipart':
        			    continue
    				# Applications should really sanitize the given filename so that an
    				# email message can't be used to overwrite important files
				filename = part.get_filename()
    				if not filename:
    				    if part.get_content_type() == "text/plain":
    					htmlBody += part.get_payload(decode=True)
    					continue
    				    else:
    					ext = mimetypes.guess_extension(part.get_content_type())
        			if not ext:
            			    # Use a generic bag-of-bits extension
            			    ext = '.bin'
        			    filename = 'part-%03d%s' % (counter, ext)
        			    
    				counter += 1
    				try:
    				    os.mkdir(os.path.join(mbName,"message%04d"%(int(num))))
    				except:
    				    pass
    				try:    
    				    attPayload = part.get_payload(decode=True)
    				    if attPayload == None:
    					continue
    				    else:
    					fp = open(os.path.join(mbName,"message%04d"%(int(num)),filename), 'wb')
    					fp.write(attPayload)
    					fp.close()
    					htmlAttachs += "<li>[%d]<a href='%s'>%s</a></li>\n"%(counter,os.path.join(mbName,"message%04d"%(int(num)), filename),filename)
    				except Exception,e:
    				    print "Failed saving %s because: %r"%("message%04d"%(int(num)),e)
    				    
    			    # Preleva dati dall'header
    			    subject = msgData.get_all('subject', [])    			    			    
    			    tos = msgData.get_all('to', [])
    			    froms = msgData.get_all('from', [])
    			    ccs = msgData.get_all('cc', [])
    			    
    			    htmlHeadField = "<b>To: %s</b><br>\n"%",".join(tos)
    			    htmlHeadField += "<b>From: %s</b><br>\n"%",".join(froms)
    			    htmlHeadField = "<b>Subject: %s</b><br>\n"%(subject[0])
    			    htmlHeadField += "<b>CCs: %s</b><br>\n"%",".join(ccs)
    			    htmlHeadField += "<b>Date: %s</b><br>\n"%datetime(msgDate[0],msgDate[1],msgDate[2],msgDate[3],msgDate[4],msgDate[5])
			    htmlHeadField += "<hr>\n"

    			    htmlHeader = "<html><head><title>%s</title></head>\n"%(subject[0])
    			    htmlFooter = "<br clear='all'><p align='center'>build by imap2mbox (C) 2008-2009 Michele Pinassi</p>\n";
    			        			
    			    # Ok, adesso salva HTML della mail...
    			    
    			    os.write(mboxFd,str(htmlHeader))
    			    os.write(mboxFd,"<body>\n")
    			    os.write(mboxFd,str(htmlHeadField))
    			    os.write(mboxFd,"<pre>"+str(htmlBody)+"</pre>")
    			    os.write(mboxFd,"<ul>"+str(htmlAttachs)+"</ul>")
    			    os.write(mboxFd,str(htmlFooter))
    			    os.write(mboxFd,"</body></html>\n")
			    os.close(mboxFd)
			    
			    htmlMexList += "<li>[%d] <a href='%s'>%s</a> from %s</li>\n"%(int(num),os.path.join(mbName,"message%04d.html"%(int(num))),subject[0],",".join(froms))
			    
		        else:
			    os.write(mboxFd,str(msgData))
			    os.write(mboxFd,"\n")
			messageSaved = messageSaved + 1
			if options.markdelete:
			    imap4.store(num, '+FLAGS', '\\Deleted')			
		    else:
			messageSkipped = messageSkipped + 1

		    print pBar, "\r",
		    pBar.updateAmount(int(num))
		    sys.stdout.flush()
		# Finito !
		print pBar, "\r",
		pBar.updateAmount(numMex)
		sys.stdout.flush()
		print "\n"
		if options.output == "html":
		    # Scrivi elenco mails
		    htmlMexList += "</ul></body></html>";
		    fp = os.open(mbName+".html", os.O_WRONLY | os.O_CREAT)
		    os.write(fp,htmlMexList)
		    os.close(fp)
		else:
		    # Chiudi file mbox
		    os.close(mboxFd)
		print "=== mbox %s have %d message saved and %d skipped ==="%(mbName,messageSaved,messageSkipped)
	else:
	    if ret == 'NO':
		print "Skip invalid mbox %s"%(mbName)	    
	    else:
		print "Skip empty mbox %s"%(mbName)
	
# Chiudi tutto !	
imap4.close()
imap4.logout()
        
